#include "geometry.h"

//#include <mex.h>

using std::vector;
using std::max;
using std::min;

using cv::Mat;

namespace ltflo {

bool tanTableSet = false;
double TAN_TABLE[1572]; /* 0..pi/2, step 1/1000, dense enough around 0, not interesting higher (angleInlierMax ~ 0.2) */

bool atanTableSet = false;
double ATAN_TABLE[10991]; /* 0..10, step 1/1000, then 10..100, step 1/10, then 100..1000, step 10, dense enough all the time (1 mrad) */
/* Precision of geom. error estimation is better than 100 ppm, maybe too precise? Speed-up: smaller tables. */

/*function [d] = lineDistSq(l1, l2, win) % lines have to be already normalized!*/
void lineDistSq(double * d, const double * l1, const double * l2, unsigned len, const double * win) {
	unsigned i;
	double dAng, distFac, angFac;
	double HALF_PI = M_PI / 2.0;
	
	/*% Lebeda method*/
	double L = 50; /*% px, magic constant, something like "quarter of mean length of intersection of line and area of interest"*/
	double L2 = 2*L;
	/*poi = sum(win,2) / 4;*/
	double poix = (win[0] + win[2] + win[4] + win[6]) / 4;
	double poiy = (win[1] + win[3] + win[5] + win[7]) / 4;
	
	if (!tanTableSet) {
		tanTableSet = 1;
		for (i = 0; i < 1572; ++i) { /* range 0..pi/2, step 0.001 */
			TAN_TABLE[i] = tan(0.001*i);
		}
	}
	if (!atanTableSet) {
		unsigned i;
		atanTableSet = 1;
		for (i = 0; i < 10000; ++i) { /* range 0..10, step 0.001 */
			ATAN_TABLE[i] = atan(0.001*i);
		}
		for (i = 100; i < 1000; ++i) { /* range 10..100, step 0.1 */
			ATAN_TABLE[10000 + i - 100] = atan(0.1 * i);
		}
		for (i = 10; i <= 100; ++i) { /* range 100..1000, step 10 */
			ATAN_TABLE[10900 + i - 10] = atan(10 * i);
		}
	}
	
	for ( i = 0; i < len; ++i ) {
		/*c = e2p(poi)' * l1 - e2p(poi)' * l2; % difference of distances to point of interest*/
		distFac = (l1[0] * poix + l1[1] * poiy + l1[2]) - (l2[0] * poix + l2[1] * poiy + l2[2]);
		
		
		/* let's try it with fast approximations! */
/*		dAng = fastInvTan1(l1[1] / l1[0]) - fastInvTan1(l2[1] / l2[0]);
		if (dAng < 0) {
			dAng += M_PI;
		}
		if ( dAng > HALF_PI ) {
			dAng = M_PI - dAng;
		}
		angFac = L2 * fastTan1(dAng / 2);
		*d = distFac * distFac + angFac * angFac;
		printf("A1: %10.2f, ", *d);
		
		dAng = lutATan(l1[1] / l1[0]) - lutATan(l2[1] / l2[0]);
		if (dAng < 0) {
			dAng += M_PI;
		}
		if ( dAng > HALF_PI ) {
			dAng = M_PI - dAng;
		}
		angFac = L2 * lutTan(dAng / 2);
		*d = distFac * distFac + angFac * angFac;
		printf("Alut: %10.2f, ", *d);
		
		dAng = lutInterpATan(l1[1] / l1[0]) - lutInterpATan(l2[1] / l2[0]);
		if (dAng < 0) {
			dAng += M_PI;
		}
		if ( dAng > HALF_PI ) {
			dAng = M_PI - dAng;
		}
		angFac = L2 * lutInterpTan(dAng / 2);
		*d = distFac * distFac + angFac * angFac;
		printf("AlutIn: %10.4f, ", *d);*/
		/* End of fast stuff */
		
		
		/*ang1 = atan(l1(2,:) ./ l1(1,:));
		ang2 = atan(l2(2,:) ./ l2(1,:));
		dAng = mod(ang1 - ang2, pi);
		dAng(dAng > pi/2) = pi - dAng(dAng > pi/2); % mutual angle
		b = 2 * L * tan(dAng / 2);*/
		
		/*dAng = atan(l1[1] / l1[0]) - atan(l2[1] / l2[0]); */
		/*dAng = doubleMod(dAng, M_PI);*/
		/*if (dAng >= M_PI) { the return value is in the range [-pi/2, pi/2] => cannot happen for the difference to be > pi
			dAng -= M_PI;
		} else */
		/* dAng = atan(l1[1] / l1[0]) - atan(l2[1] / l2[0]);// -> slow precise computation
		if (dAng < 0) {
			dAng += M_PI;
		}
		if ( dAng > HALF_PI ) {
			dAng = M_PI - dAng;
		}
		angFac = L2 * tan(dAng / 2);// -> end of slow precise computation */
		
		dAng = lutATanS(l1[1] / l1[0]) - lutATanS(l2[1] / l2[0]);
		if (dAng < 0) {
			dAng += M_PI;
		}
		if ( dAng > HALF_PI ) {
			dAng = M_PI - dAng;
		}
		/* we know the angle is in range [0;pi/2] and the TAN_TABLE is already set */
		angFac = L2 * TAN_TABLE[(unsigned)(dAng*500)]; /* dAng/2, *1000 for index */
		
		*d = distFac * distFac + angFac * angFac;
		/*printf("R: %10.4f.\n", *d);*/
		
		l1 += 3;
		l2 += 3;
		++d;
	}
}

/*function [d] = aLineDistSq(a, l1, l2, win)*/
void aLineDistSq(double * d, const double * A, const double * l1, const double * l2, unsigned len, const double * win, double *buffer) {
	double * l1p, * l2p, * d2, * dd = 0, win3[3*4], win3p[3*4];
	unsigned i;
	
	if (buffer) {
		l1p = buffer;
		l2p = buffer + 3 * len;
		d2 = buffer + 6 * len;
	} else {
		l1p = (double *) malloc(3*len*sizeof(double));
		l2p = (double *) malloc(3*len*sizeof(double));
		dd = d2 = (double *) malloc(len*sizeof(double));
	}
	
	/*l2p = normLine(a' \ l1);*/
	aTInvMultiply(l2p, A, l1, len);
	normLines(l2p, len);
	/*l1p = normLine(a' * l2);*/
	aTMultiply(l1p, A, l2, len);
	normLines(l1p, len);
	
	/*d = lineDistSq(l1, l1p, win) + lineDistSq(l2, l2p, p2e(a*e2p(win)));*/
	lineDistSq(d, l1, l1p, len, win);
	/*p2e(a*e2p(win))*/
	for (i = 0; i < 4; ++i) {
		win3[i*3 + 0] = win[i*2 + 0];
		win3[i*3 + 1] = win[i*2 + 1];
		win3[i*3 + 2] = 1;
	}
	aMultiply(win3p, A, win3, 4);
	for (i = 0; i < 4; ++i) { /* This is a really nasty stuff and I should be executed for writing it, but they really won't collide :) */
		win3p[i*2 + 0] = win3p[i*3 + 0] / win3p[i*3 + 2];
		win3p[i*2 + 1] = win3p[i*3 + 1] / win3p[i*3 + 2];
	}
	lineDistSq(d2, l2, l2p, len, win3p);
	for ( i = 0; i < len; ++i ) {
		*d++ += *d2++;
	}

	if (!buffer) {
		free(l1p);
		free(l2p);
		free(dd);
	}
}

/*function [d] = aError(a, u(:,inliers))
	returned 'd' is of size 4xLEN!
	buffer has to be sized 24*LEN
	pass NULL in 'inliers' if all the points are measured */
void aErrorU(double * d, const double * A, const double * u, const unsigned * inliers, unsigned len, unsigned uLen, double * buffer) {
	double * l1, * l1p, * l2, * l2p, * u1, * u1p, * u2, * u2p;
	const double * ui;
	unsigned i;
	if ( buffer ) {
		l1  = buffer;
		l1p = buffer +  3 * len;
		l2  = buffer +  6 * len;
		l2p = buffer +  9 * len;
		u1  = buffer + 12 * len;
		u1p = buffer + 15 * len;
		u2  = buffer + 18 * len;
		u2p = buffer + 21 * len;
	} else {
		l1  = (double *) malloc(3 * len * sizeof(double));
		l1p = (double *) malloc(3 * len * sizeof(double));
		l2  = (double *) malloc(3 * len * sizeof(double));
		l2p = (double *) malloc(3 * len * sizeof(double));
		u1  = (double *) malloc(3 * len * sizeof(double));
		u1p = (double *) malloc(3 * len * sizeof(double));
		u2  = (double *) malloc(3 * len * sizeof(double));
		u2p = (double *) malloc(3 * len * sizeof(double));
	}
	
	for ( i = 0; i < len; ++i ) {
	// u is in form in form [l1;l2;x1;x2;ind]
		if ( inliers ) {
			ui = u + inliers[i] * uLen;
		} else {
			ui = u + i * uLen;
		}
		*l1++ = *ui++;
		*l1++ = *ui++;
		*l1++ = *ui++;
		*l2++ = *ui++;
		*l2++ = *ui++;
		*l2++ = *ui++;
		*u1++ = *ui++;
		*u1++ = *ui++;
		*u1++ = *ui++;
		*u2++ = *ui++;
		*u2++ = *ui++;
		*u2++ = *ui++;
	}
	l1 -= 3 * len;
	l2 -= 3 * len;
	u1 -= 3 * len;
	u2 -= 3 * len;
	
	/* l2p = normLine(A' \ l1); */
	aTInvMultiply(l2p, A, l1, len);
	normLines(l2p, len);
	/* l1p = normLine(a' * l2); */
	aTMultiply(l1p, A, l2, len);
	normLines(l1p, len);
	/* u2p = e2p(p2e(A * u1)); */
	aMultiply(u2p, A, u1, len);
	normPoints(u2p, len);
	/* u1p = e2p(p2e(A \ u2)); */
	aInvMultiply(u1p, A, u2, len);
	normPoints(u1p, len);
	
	for ( i = 0; i < len; ++i ) {
		*d++ = u1[0] * l1p[0] + u1[1] * l1p[1] + u1[2] * l1p[2];
		*d++ = u2[0] * l2p[0] + u2[1] * l2p[1] + u2[2] * l2p[2];
		*d++ = u1p[0] * l1[0] + u1p[1] * l1[1] + u1p[2] * l1[2];
		*d++ = u2p[0] * l2[0] + u2p[1] * l2[1] + u2p[2] * l2[2];
		l1 += 3; l1p += 3; l2 += 3; l2p += 3; u1 += 3; u1p += 3; u2 += 3; u2p += 3;
	}
	
	if ( !buffer ) {
		free(l1 - 3 * len);
		free(l1p- 3 * len);
		free(l2 - 3 * len);
		free(l2p- 3 * len);
		free(u1 - 3 * len);
		free(u1p- 3 * len);
		free(u2 - 3 * len);
		free(u2p- 3 * len);
	}
}
void aErrorSqU(double * d, const double * A, const double * u, const unsigned * inliers, unsigned len, unsigned uLen, double * buffer) {
	double * d4 = new double[4*len];
	aErrorU(d4, A, u, inliers, len, uLen, buffer);
	for (unsigned i = 0; i < len; ++i) {
		*d++ = d4[0]*d4[0] + d4[1]*d4[1] + d4[2]*d4[2] + d4[3]*d4[3];
		d4 += 4;
	}
	delete[] (d4 - 4*len);
}

/*function [d] = aError(a, u(:,inliers))
	returned 'd' is of size 4xLEN!
	buffer has to be sized 12*LEN */
void aError(double * d, const double * A, const double * l1, const double * l2, const double * u1, const double * u2, unsigned len, double * buffer) {
	double * l1p, * l2p, * u1p, * u2p;
	unsigned i;
	if ( buffer ) {
		l1p = buffer + 0 * len;
		l2p = buffer + 3 * len;
		u1p = buffer + 6 * len;
		u2p = buffer + 9 * len;
	} else {
		l1p = (double *) malloc(3 * len * sizeof(double));
		l2p = (double *) malloc(3 * len * sizeof(double));
		u1p = (double *) malloc(3 * len * sizeof(double));
		u2p = (double *) malloc(3 * len * sizeof(double));
	}
	
	/* l2p = normLine(A' \ l1); */
	aTInvMultiply(l2p, A, l1, len);
	normLines(l2p, len);
	/* l1p = normLine(a' * l2); */
	aTMultiply(l1p, A, l2, len);
	normLines(l1p, len);
	/* u2p = e2p(p2e(A * u1)); */
	aMultiply(u2p, A, u1, len);
	normPoints(u2p, len);
	/* u1p = e2p(p2e(A \ u2)); */
	aInvMultiply(u1p, A, u2, len);
	normPoints(u1p, len);
	
	for ( i = 0; i < len; ++i ) {
		*d++ = u1[0] * l1p[0] + u1[1] * l1p[1] + u1[2] * l1p[2];
		*d++ = u2[0] * l2p[0] + u2[1] * l2p[1] + u2[2] * l2p[2];
		*d++ = u1p[0] * l1[0] + u1p[1] * l1[1] + u1p[2] * l1[2];
		*d++ = u2p[0] * l2[0] + u2p[1] * l2[1] + u2p[2] * l2[2];
		l1 += 3; l1p += 3; l2 += 3; l2p += 3; u1 += 3; u1p += 3; u2 += 3; u2p += 3;
	}
	
	if ( !buffer ) {
		free(l1p - 3 * len);
		free(l2p - 3 * len);
		free(u1p - 3 * len);
		free(u2p - 3 * len);
	}
}

/*l2 = inv(a^T) l1*/
void aTInvMultiply(double * l2, const double * a, const double * l1, unsigned len) {
	double ai[3*3]; /*inverted*/
	double denom = det3(a); /*a[0]*a[4]*a[8]-a[0]*a[5]*a[7]-a[1]*a[3]*a[8]+a[1]*a[5]*a[6]+a[2]*a[3]*a[7]-a[2]*a[4]*a[6];*/
	ai[0] = (a[4]*a[8]-a[5]*a[7]) / denom;
	ai[1] = (a[2]*a[7]-a[1]*a[8]) / denom;
	ai[2] = (a[1]*a[5]-a[2]*a[4]) / denom;
	ai[3] = (a[5]*a[6]-a[3]*a[8]) / denom;
	ai[4] = (a[0]*a[8]-a[2]*a[6]) / denom;
	ai[5] = (a[2]*a[3]-a[0]*a[5]) / denom;
	ai[6] = (a[3]*a[7]-a[4]*a[6]) / denom;
	ai[7] = (a[1]*a[6]-a[0]*a[7]) / denom;
	ai[8] = (a[0]*a[4]-a[1]*a[3]) / denom;
	aTMultiply(l2, ai, l1, len);
}

/*l2 = inv(a) l1*/
void aInvMultiply(double * l2, const double * a, const double * l1, unsigned len) {
	double ai[3*3]; /*inverted*/
	double denom = det3(a); /*a[0]*a[4]*a[8]-a[0]*a[5]*a[7]-a[1]*a[3]*a[8]+a[1]*a[5]*a[6]+a[2]*a[3]*a[7]-a[2]*a[4]*a[6];*/
	ai[0] = (a[4]*a[8]-a[5]*a[7]) / denom;
	ai[1] = (a[2]*a[7]-a[1]*a[8]) / denom;
	ai[2] = (a[1]*a[5]-a[2]*a[4]) / denom;
	ai[3] = (a[5]*a[6]-a[3]*a[8]) / denom;
	ai[4] = (a[0]*a[8]-a[2]*a[6]) / denom;
	ai[5] = (a[2]*a[3]-a[0]*a[5]) / denom;
	ai[6] = (a[3]*a[7]-a[4]*a[6]) / denom;
	ai[7] = (a[1]*a[6]-a[0]*a[7]) / denom;
	ai[8] = (a[0]*a[4]-a[1]*a[3]) / denom;
	aMultiply(l2, ai, l1, len);
}

/*l2 = a^T l1*/
void aTMultiply(double * l2, const double * a, const double * l1, unsigned len) {
	unsigned i;
	double k1, k2, k3;
	for ( i = 0; i < len; ++i ) {
		k1 = *l1++;
		k2 = *l1++;
		k3 = *l1++;
		*l2++ = a[0]*k1 + a[1]*k2 + a[2]*k3;
		*l2++ = a[3]*k1 + a[4]*k2 + a[5]*k3;
		*l2++ = a[6]*k1 + a[7]*k2 + a[8]*k3;
	}
}

/*l2 = a l1*/
void aMultiply(double * l2, const double * a, const double * l1, unsigned len) {
	unsigned i;
	for ( i = 0; i < len; ++i ) {
		*l2++ = a[0]*l1[0] + a[3]*l1[1] + a[6]*l1[2];
		*l2++ = a[1]*l1[0] + a[4]*l1[1] + a[7]*l1[2];
		*l2++ = a[2]*l1[0] + a[5]*l1[1] + a[8]*l1[2];
		l1 += 3;
	}
}

void projectA(double *xp, double *yp, const double *A, const unsigned *x, const unsigned *y, unsigned LEN) {
	unsigned i;
	for (i = 0; i < LEN; ++i) {
		*xp++ = A[0] * *x + A[3] * *y + A[6];
		*yp++ = A[1] * *x + A[4] * *y + A[7];
		++x;
		++y;
	}
}

// this one does the projection and additionally creates the "correspFull" structure with original points as well, and rounds the result
std::vector<unsigned> projectCorrespondences(const cv::Mat& A, const std::vector<cv::Point>& points, cv::Size limits) {
	vector<unsigned> correspFull;
	correspFull.reserve(4 * points.size());
	for (vector<cv::Point>::const_iterator it = points.begin(); it != points.end(); ++it) {
		cv::Point p = *it;
		correspFull.push_back(p.x);
		correspFull.push_back(p.y);
		int xp = round(A.at<double>(0,0) * p.x + A.at<double>(0,1) * p.y + A.at<double>(0,2));
		correspFull.push_back(max(0, min(limits.width-1, xp)));
		int yp = round(A.at<double>(1,0) * p.x + A.at<double>(1,1) * p.y + A.at<double>(1,2));
		correspFull.push_back(max(0, min(limits.height-1, yp)));
	}
	
	return correspFull;
}


void projectInvA(double *xp, double *yp, const double *a, const unsigned *x, const unsigned *y, unsigned LEN) {
	double ai[3*3]; /*inverted*/
	double denom = det3(a); /*a[0]*a[4]*a[8]-a[0]*a[5]*a[7]-a[1]*a[3]*a[8]+a[1]*a[5]*a[6]+a[2]*a[3]*a[7]-a[2]*a[4]*a[6];*/
	ai[0] = (a[4]*a[8]-a[5]*a[7]) / denom;
	ai[1] = (a[2]*a[7]-a[1]*a[8]) / denom;
	ai[2] = (a[1]*a[5]-a[2]*a[4]) / denom;
	ai[3] = (a[5]*a[6]-a[3]*a[8]) / denom;
	ai[4] = (a[0]*a[8]-a[2]*a[6]) / denom;
	ai[5] = (a[2]*a[3]-a[0]*a[5]) / denom;
	ai[6] = (a[3]*a[7]-a[4]*a[6]) / denom;
	ai[7] = (a[1]*a[6]-a[0]*a[7]) / denom;
	ai[8] = (a[0]*a[4]-a[1]*a[3]) / denom;
	projectA(xp, yp, ai, x, y, LEN);
}

/*Normalize to unitary length on normal vector*/
void normLines(double * l, unsigned len) {
	unsigned i;
	double vl; /*vector length*/
	for ( i = 0; i < len; ++i ) {
		vl = sqrt(l[0]*l[0] + l[1]*l[1]);
		*l++ /= vl;
		*l++ /= vl;
		*l++ /= vl;
	}
}

/*Normalize to unitary third coordinate*/
void normPoints(double * u, unsigned len) {
	unsigned i;
	double third;
	for ( i = 0; i < len; ++i ) {
		third = u[2];
		*u++ /= third;
		*u++ /= third;
		*u++ /= third;
	}
}

/*aLineDistSq with compact correspondence input, just wrapper*/
void aLineDistSqU(double * d, const double * A, const double * u, unsigned len, const double *win, unsigned correspLength, double *buffer) {
	double *l1, *l2, *buff2;
	unsigned i;
	
	if (buffer) {
		l1 = buffer;
		l2 = buffer + 3 * len;
		buff2 = buffer + 6 * len;
	} else {
		l1 = (double *) malloc(3 * len * sizeof(double));
		l2 = (double *) malloc(3 * len * sizeof(double));
		buff2 = 0;
	}
	
	for (i = 0; i < len; ++i) { /* speed-up: don't use this function :) */
		*l1++ = *u++;
		*l1++ = *u++;
		*l1++ = *u++;
		*l2++ = *u++;
		*l2++ = *u++;
		*l2++ = *u++;
		u += correspLength - 6;
	}
	
	l1 -= 3*len;
	l2 -= 3*len;
	
	aLineDistSq(d, A, l1, l2, len, win, buff2);
	
	if (!buffer) {
		free(l1);
		free(l2);
	}
}

/*Area of a rectangle*/
double rectangleArea(point * rect) {
	double a,b;
	a = sqrt((rect[0].x - rect[1].x) * (rect[0].x - rect[1].x) + (rect[0].y - rect[1].y) * (rect[0].y - rect[1].y));
	b = sqrt((rect[2].x - rect[1].x) * (rect[2].x - rect[1].x) + (rect[2].y - rect[1].y) * (rect[2].y - rect[1].y));
	return a * b;
}


/*computes sum of squared geometric errors of line correspondences - distance-angle measure*/
void geomError(double *p, double *hx, int m, int n, void *adata) {
	arr6 * a = (arr6 *) adata;
	double A[3*3];
	unsigned i;
	
	A[0] = A[4] = p[0] / 100;
	A[1] = p[1] / 100;
	A[2] = A[5] = 0;
	A[3] = -p[1] / 100;
	A[6] = p[2];
	A[7] = p[3];
	A[8] = 1;
	
	aLineDistSq(a->d, A, a->l1, a->l2, a->size, a->win, a->buffer);
	
	*hx = 0; /* parametrisation? */
	for (i = 0; i < a->size; ++i) {
		*hx += a->d[i];
	}
	
}

/*computes sum of squared geometric errors of line correspondences - distance-angle measure*/
double geomErrorSimple(double *p, void *adata) {
	arr6 * a = (arr6 *) adata;
	double A[3*3], val;
	unsigned i;
	
	A[0] = A[4] = p[0] / 100;
	A[1] = p[1] / 100;
	A[2] = A[5] = 0;
	A[3] = -p[1] / 100;
	A[6] = p[2];
	A[7] = p[3];
	A[8] = 1;
	
	aLineDistSq(a->d, A, a->l1, a->l2, a->size, a->win, a->buffer);
	
	val = 0;
	for (i = 0; i < a->size; ++i) {
		val += a->d[i];
	}
	return val;
}

/*computes sum of squared geometric errors of line correspondences - four line-point distances
	buffer has to be sized 16*LEN */
double geomErrorSimple4D(double * p, void * adata) {
	arr4D * a = (arr4D *) adata;
	double A[3*3], val;
	unsigned i;
	
	A[0] = A[4] = p[0];
	A[1] = p[1];
	A[2] = A[5] = 0;
	A[3] = -p[1];
	A[6] = p[2];
	A[7] = p[3];
	A[8] = 1;
	
	aError(a->buffer, A, a->l1, a->l2, a->u1, a->u2, a->size, a->buffer + 4*a->size);
	
	val = 0;
	for (i = 0; i < a->size * 4; ++i) {
		val += a->buffer[i] * a->buffer[i];
	}
	return val;
}

///*computes sum of squared geometric errors of line correspondences
//	buffer has to be sized 12*LEN*/
//vector<double> LineErrorF::operator()(const vector<double>& p) const {
//	double A[3*3];
//	vector<double> vec(4 * size);
//	
//	A[0] = A[4] = p[0];
//	A[1] = p[1];
//	A[2] = A[5] = 0;
//	A[3] = -p[1];
//	A[6] = p[2];
//	A[7] = p[3];
//	A[8] = 1;
//	
//	aError(&vec[0], A, l1, l2, u1, u2, size, buffer2);
//	
//	return vec;
//}

/* computes (separated) geometric errors of line correspondences (orig. error according to ACCV)
	buffer has to be sized 6*LEN */
//vector<double> LineErrorOrigF::operator()(const vector<double>& p) const {
//	double A[3*3], win3[3*4], win3p[3*4];
//	vector<double> vecRes(4 * size);
//	double * l1p, * l2p;
//	double * d, * d2;
//	unsigned i;
//	
//	A[0] = A[4] = p[0];
//	A[1] = p[1];
//	A[2] = A[5] = 0;
//	A[3] = -p[1];
//	A[6] = p[2];
//	A[7] = p[3];
//	A[8] = 1;
//	
//	l1p = buffer2;
//	l2p = buffer2 + 3 * size;
//	d = &vecRes[0];
//	d2 = d + 1; // we return it as A1_1,A2_1,P1_1,P2_1,...A1_N,A2_N,P1_N,P2_N, so the second image has it shifted by 1
//	
//	/*l2p = normLine(a' \ l1);*/
//	aTInvMultiply(l2p, A, l1, size);
//	normLines(l2p, size);
//	/*l1p = normLine(a' * l2);*/
//	aTMultiply(l1p, A, l2, size);
//	normLines(l1p, size);
//	
//	/*d = lineDistSq(l1, l1p, win) + lineDistSq(l2, l2p, p2e(a*e2p(win))); - but separated!*/
//	lineDistSeparated(d, l1, l1p, size, win);
//	/*p2e(a*e2p(win))*/
//	for (i = 0; i < 4; ++i) {
//		win3[i*3 + 0] = win[i*2 + 0];
//		win3[i*3 + 1] = win[i*2 + 1];
//		win3[i*3 + 2] = 1;
//	}
//	aMultiply(win3p, A, win3, 4);
//	for (i = 0; i < 4; ++i) { /* This is some really nasty stuff, but they really won't collide :) */
//		win3p[i*2 + 0] = win3p[i*3 + 0] / win3p[i*3 + 2];
//		win3p[i*2 + 1] = win3p[i*3 + 1] / win3p[i*3 + 2];
//	}
//	lineDistSeparated(d2, l2, l2p, size, win3p);
//	
//	return vecRes;
//}

/* Separated components of the original geometric error
	Lines have to be already normalized! */
void lineDistSeparated(double * d, const double * l1, const double * l2, unsigned len, const double * win) {
	unsigned i;
	double dAng, distFac, angFac;
	double HALF_PI = M_PI / 2.0;
	
	/*% Lebeda method*/
	double L = 50; /*% px, magic constant, something like "quarter of mean length of intersection of line and area of interest"*/
	double L2 = 2*L;
	/*poi = sum(win,2) / 4;*/
	double poix = (win[0] + win[2] + win[4] + win[6]) / 4;
	double poiy = (win[1] + win[3] + win[5] + win[7]) / 4;
	
	if (!tanTableSet) {
		tanTableSet = 1;
		for (i = 0; i < 1572; ++i) { /* range 0..pi/2, step 0.001 */
			TAN_TABLE[i] = tan(0.001*i);
		}
	}
	if (!atanTableSet) {
		unsigned i;
		atanTableSet = 1;
		for (i = 0; i < 10000; ++i) { /* range 0..10, step 0.001 */
			ATAN_TABLE[i] = atan(0.001*i);
		}
		for (i = 100; i < 1000; ++i) { /* range 10..100, step 0.1 */
			ATAN_TABLE[10000 + i - 100] = atan(0.1 * i);
		}
		for (i = 10; i <= 100; ++i) { /* range 100..1000, step 10 */
			ATAN_TABLE[10900 + i - 10] = atan(10 * i);
		}
	}
	
	for ( i = 0; i < len; ++i ) {
		/*c = e2p(poi)' * l1 - e2p(poi)' * l2; % difference of distances to point of interest*/
		distFac = (l1[0] * poix + l1[1] * poiy + l1[2]) - (l2[0] * poix + l2[1] * poiy + l2[2]);
		
		/*ang1 = atan(l1(2,:) ./ l1(1,:));
		ang2 = atan(l2(2,:) ./ l2(1,:));
		dAng = mod(ang1 - ang2, pi);
		dAng(dAng > pi/2) = pi - dAng(dAng > pi/2); % mutual angle
		b = 2 * L * tan(dAng / 2);*/
		
//		dAng = atan(l1[1] / l1[0]) - atan(l2[1] / l2[0]); // -> slow precise computation
		dAng = lutInterpATanS(l1[1] / l1[0]) - lutInterpATanS(l2[1] / l2[0]); // -> LUT with interpolation
		if (dAng < 0) {
			dAng += M_PI;
		}
		if ( dAng > HALF_PI ) {
			dAng = M_PI - dAng;
		}
//		angFac = L2 * tan(dAng / 2);// -> slow precise computation */
		angFac = L2 * lutInterpTanS(dAng / 2);// -> LUT with interpolation again
		
		*d = abs(angFac);
		d += 2;
		*d = abs(distFac);
		d += 2; // we return it as A1_1,A2_1,P1_1,P2_1,...A1_N,A2_N,P1_N,P2_N, so the distFac is offset by 2 and there are spaces between each.
				// The second image is then called offset by 1, so the spaces are filled in the second run
		
		l1 += 3;
		l2 += 3;
	}
}

///* computes gradients of geometric errors of line correspondences (orig. error according to ACCV) */
//int LineErrorOrigG::operator()(const vector<double>& p, cv::Mat & gradients, int basePtr) const {
////	vector<vector<double> > vecRes(4*size,vector<double>(4,0));
//	
//	double * d = ((double *)gradients.data) + basePtr;
//	
//	unsigned i;
//	double L1x,L1y,L1z,L2x,L2y,L2z,A1,A2,A3,A4,L,wx,wy;
//	double v000, v001, v002, v003, v004, v005, v006, v007, v008, v009, v010, v011, v012, v013, v014, v015, v016, v017, v018, v019;
//	double v020, v021, v022, v023, v024, v025, v026, v027, v028, v029, v030, v031, v032, v033, v034, v035, v036, v037, v038, v039;
//	double v040, v041, v042, v043, v044, v045, v046, v047, v048, v049, v050, v051, v052;
//	double v099, v098;
//	
//	A1 = p[0]; A2 = p[1]; A3 = p[2]; A4 = p[3];
//	L = 50;
//	wx = (win[0] + win[2] + win[4] + win[7]) / 4;
//	wy = (win[1] + win[3] + win[5] + win[7]) / 4;
//	
//	v000 = A2 * A2;
//	v001 = A1 * A1;
//	v007 = L*L;
//	v013 = v000 + v001;
//	v018 = 1/v013;
//	
//	for ( i = 0; i < size; ++i ) {
//		L1x = l1[i * 3 + 0];
//		L1y = l1[i * 3 + 1];
//		L1z = l1[i * 3 + 2];
//		L2x = l2[i * 3 + 0];
//		L2y = l2[i * 3 + 1];
//		L2z = l2[i * 3 + 2];
//		
//		v002 = L2x * wy;
//		v003 = L2y * wy;
//		v004 = L2x * wx;
//		v005 = A4 * L2y;
//		v006 = A3 * L2x;
//		v008 = A1 * v003;
//		v009 = A2 * L2y * wx;
//		v010 = A1 * v004;
//		v011 = -(A2 * v002);
//		v012 = L1x*L1x + L1y*L1y;
//		v014 = L2x*L2x + L2y*L2y;
//		v015 = v003 + v004;
//		v016 = L1z + L1x * wx + L1y * wy;
//		v017 = v002 - L2y * wx;
//		v019 = 1/v014;
//		v020 = 1/v012;
//		v021 = v012*v012;
//		v022 = sqrt(v014);
//		v023 = sqrt(v012);
//		v024 = sqrt(v012*v012*v012);
//		v025 = sqrt(v014*v014*v014);
//		v026 = v013 * v014;
//		v027 = v012 * v018;
//		v028 = A1 * L1x * L2x - A2 * L1y * L2x + A2 * L1x * L2y + A1 * L1y * L2y;
//		v029 = A2 * L1x * L2x + A1 * L1y * L2x - A1 * L1x * L2y + A2 * L1y * L2y;
//		v030 = L2z + v008 + v009 + v010 + v011;
//		v031 = 1/sqrt(v026);
//		v032 = 1/(v029*v029*v029);
//		v033 = 1/v029;
//		v034 = sqrt(v027);
//		v035 = -(v016 * v022);
//		v036 = 1/sqrt(v027);
//		v037 = v005 * v034;
//		v038 = v006 * v034;
//		v039 = v030 * v034;
//		v040 = (v020 * v028 * v034)/sqrt(v014);
//		v041 = (-1 + v040)*(-1 + v040);
//		v098 = lutInterpTan(acos((v028 * v031)/sqrt(v012))/2.);
//		v042 = sqrt(v007 * v098 * v098);
//		v099 = lutInterpTan(acos(v040)/2.);
//		v043 = 1/sqrt(v007 * v099 * v099);
//		v044 = v005 * v023 + v006 * v023 - v016 * sqrt(v026) + v023 * v030;
//		v045 = 1/v044;
//		v046 = 1/(v044*v044*v044);
//		v047 = v035 + v037 + v038 + v039;
//		v048 = v018 * v019 * v020 * v044 * v044;
//		v049 = sqrt(v048);
//		v050 = sqrt(v048*v048*v048);
//		v051 = v013 * v019 * v020 * v047 * v047;
//		v052 = 1/sqrt(v051);
//		
//		// gradients vector is a long vector of short vectors, so we fill 4 rows of it, each with gradients for each variable according to one error of one correspondence (in order A1,A2,P1,P2)
//		
//		// A1
////		sizes[0] = i*4 + 0;
////		sizes[1] = 0;
//		*d++ = -2 * A2 * v014 * v023 * v031 * v033 * v042; // d A1 / d p[0]
////		vecRes[i*4 + 0][0] = -2 * A2 * v014 * v023 * v031 * v033 * v042; // d A1 / d p[0]
////		sizes[1] = 1;
//		*d++ = 2 * A1 * v014 * v023 * v031 * v033 * v042; // d A1 / d p[1]
////		vecRes[i*4 + 0][1] = 2 * A1 * v014 * v023 * v031 * v033 * v042; // d A1 / d p[1]
//		*d++ = 0;
//		*d++ = 0;
//		//vecRes[i*4 + 0][2] = 0; // d A1 / d p[2] // zeros are preset
//		//vecRes[i*4 + 0][3] = 0; // d A1 / d p[3]
////		sizes[1] = 2;
////		printf("\n%e\n",gradTens[sizes]);
//		
//		// A2
////		sizes[0] = i*4 + 1;
////		sizes[1] = 0;
//		*d++ = -2 * A2 * v007 * v021 * v025 * v032 * v036 * v041 * v043; // d A2 / d p[0]
////		vecRes[i*4 + 1][0] = -2 * A2 * v007 * v021 * v025 * v032 * v036 * v041 * v043; // d A2 / d p[0]
////		sizes[1] = 1;
//		*d++ = 2 * A1 * v007 * v021 * v025 * v032 * v036 * v041 * v043; // d A2 / d p[1]
////		vecRes[i*4 + 1][1] = 2 * A1 * v007 * v021 * v025 * v032 * v036 * v041 * v043; // d A2 / d p[1]
//		*d++ = 0;
//		*d++ = 0;
//		//vecRes[i*4 + 1][2] = 0; // d A2 / d p[2] // zeros are preset
//		//vecRes[i*4 + 1][3] = 0; // d A2 / d p[3]
//		
//		// P1
////		sizes[0] = i*4 + 2;
////		sizes[1] = 0;
//		*d++ = v014 * (-(A1 * (L2z + v005 + v006 + v009 + v011)) + v000 * v015) * v024 * v046 * v050; // d P1 / d p[0]
////		vecRes[i*4 + 2][0] = v014 * (-(A1 * (L2z + v005 + v006 + v009 + v011)) + v000 * v015) * v024 * v046 * v050; // d P1 / d p[0]
////		sizes[1] = 1;
//		*d++ = -(v014 * (A2 * (L2z + v005 + v006 + v008 + v010) + v001 * v017) * v024 * v046 * v050); // d P1 / d p[1]
////		vecRes[i*4 + 2][1] = -(v014 * (A2 * (L2z + v005 + v006 + v008 + v010) + v001 * v017) * v024 * v046 * v050); // d P1 / d p[1]
////		sizes[1] = 2;
//		*d++ = L2x * v023 * v045 * v049; // d P1 / d p[2]
////		vecRes[i*4 + 2][2] = L2x * v023 * v045 * v049; // d P1 / d p[2]
////		sizes[1] = 3;
//		*d++ = L2y * v023 * v045 * v049; // d P1 / d p[3]
////		vecRes[i*4 + 2][3] = L2y * v023 * v045 * v049; // d P1 / d p[3]
//		
//		// P2
////		sizes[0] = i*4 + 3;
////		sizes[1] = 0;
//		*d++ = (v018 * (v000 * v015 * v034 + v001 * v015 * v034 + A1 * v035) * sqrt(v051)) / v047; // d P2 / d p[0]
////		vecRes[i*4 + 2][0] = (v018 * (v000 * v015 * v034 + v001 * v015 * v034 + A1 * v035) * sqrt(v051)) / v047; // d P2 / d p[0]
////		sizes[1] = 1;
//		*d++ = v019 * v020 * (A2 * v016 * v022 + v000 * v017 * v034 + v001 * v017 * v034) * (v016 * v022 - v037 - v038 - v039) * v052; // d P2 / d p[1]
////		vecRes[i*4 + 2][1] = v019 * v020 * (A2 * v016 * v022 + v000 * v017 * v034 + v001 * v017 * v034) * (v016 * v022 - v037 - v038 - v039) * v052; // d P2 / d p[1]
////		sizes[1] = 2;
//		*d++ = L2x * v019 * v036 * v047 * v052; // d P2 / d p[2]
////		vecRes[i*4 + 2][2] = L2x * v019 * v036 * v047 * v052; // d P2 / d p[2]
////		sizes[1] = 3;
//		*d++ = L2y * v019 * v036 * v047 * v052; // d P2 / d p[3]
////		vecRes[i*4 + 2][3] = L2y * v019 * v036 * v047 * v052; // d P2 / d p[3]
//		
//	}
//	
//	return size * 4;
//}

///* computes image evidence score */
///* Basically the same as "evidenceScore", but without regularisations (local search only anyway) */
//vector<double> EvidF::operator()(const vector<double>& p) const {
//	double A[3*3];
//	vector<double> vecOut(2 * size); // residuals: location and orientation error separately
//	unsigned i, bi;
//	double x, y, w, dx, dy;
//	unsigned ux, uy;
//	double wAng, wLoc, rot;
//	const double * ui;
//	
//	A[0] = A[4] = p[0];
//	A[1] = p[1];
//	A[2] = A[5] = 0;
//	A[3] = -p[1];
//	A[6] = p[2];
//	A[7] = p[3];
//	A[8] = 1;
//	
//	rot = a2rot(A);
//	
//	for (i = 0; i < size; ++i) {
//		ui = u + inliers[i] * U_SIZE;
//		x = A[0]*ui[6] + A[3]*ui[7] + A[6]*ui[8];
//		y = A[1]*ui[6] + A[4]*ui[7] + A[7]*ui[8];
//		w = A[2]*ui[6] + A[5]*ui[7] + A[8]*ui[8];
//		x = max(1., min(N-.001, x / w)) - 1; /*coordinates to indices!*/
//		y = max(1., min(M-.001, y / w)) - 1; /* Shifted so we are never exactly on the right or bottom border */
//		bi = (unsigned)(y+.5) + M * (unsigned)(x+.5); // fast round
//		if (repair) {
//			wAng = fabs(cos(atan2(ui[1], ui[0]) + rot - evidAng[bi]));
//		} else {
//			wAng = .5 + .5 * cos(atan2(ui[1], ui[0]) + rot - evidAng[bi]);
//		}
//		// interp2 for location
//		wLoc = 0;
//		ux = unsigned(x);
//		uy = unsigned(y);
//		dx = x - ux;
//		dy = y - uy;
//		wLoc += evidDist[uy + M*ux] * (1-dx) * (1-dy);
//		wLoc += evidDist[uy + M*ux + M] * dx * (1-dy);// to the left
//		wLoc += evidDist[uy + 1 + M*ux] * (1-dx) * dy;// down
//		wLoc += evidDist[uy + 1 + M*ux + M] * dx * dy;// to the left down
//		
//		vecOut[i*2+0] = 1 - wLoc;
//		vecOut[i*2+1] = 1 - wAng;
//	}
//	
//	return vecOut;
//}


/* Computes negative sum of weights at backprojected positions */
double refineFitError(double *p, void *adata) {
	arr5 *a = (arr5 *) adata;
	double A[3*3], val = 0;
	const double *u = a->u2;
	double x, y, w;
	unsigned ux, uy, i;
	
	A[0] = A[4] = p[0];
	A[1] = p[1];
	A[2] = A[5] = 0;
	A[3] = -p[1];
	A[6] = p[2];
	A[7] = p[3];
	A[8] = 1;
	
	inv3(A);
	
	for (i = 0; i < a->LEN; ++i) {
		x = A[0]*u[0] + A[3]*u[1] + A[6]; /* speed-up: assume it is Affine, or Similarity and use it */
		y = A[1]*u[0] + A[4]*u[1] + A[7];
		w = A[2]*u[0] + A[5]*u[1] + A[8];
		ux = (unsigned) (x/w+.5); /* fast rounding */
		uy = (unsigned) (y/w+.5); /* fast rounding */
		ux = max(1u, min(a->N, ux));
		uy = max(1u, min(a->M, uy));
		val += a->wf[(ux - 1) * a->M + uy - 1];
		u += 2;
	}
	
	return - val/a->LEN;
}

// the same in the object-oriented style
double WeightsF::calc(const double* p) const {
	double val = 0;
	
	cv::Mat A = cv::Mat::eye(3, 3, CV_64F);
	A.at<double>(0, 0) = A.at<double>(1, 1) = p[0];
	A.at<double>(1, 0) = p[1];
	A.at<double>(0, 1) = -p[1];
	A.at<double>(0, 2) = p[2];
	A.at<double>(1, 2) = p[3];
	
	A = A.inv();
	
	for (unsigned i = 0; i < u2.cols; ++i) {
		double x = A.at<double>(0,0)*u2.at<double>(0,i) + A.at<double>(0,1)*u2.at<double>(1,i) + A.at<double>(0,2);
		double y = A.at<double>(1,0)*u2.at<double>(0,i) + A.at<double>(1,1)*u2.at<double>(1,i) + A.at<double>(1,2);
		unsigned ux = (unsigned) (x+.5); /* fast rounding */
		unsigned uy = (unsigned) (y+.5); /* fast rounding */
		ux = max(0u, min(wf.cols - 1u, ux));
		uy = max(0u, min(wf.rows - 1u, uy));
		val += wf.at<double>(uy, ux);
	}
	return -val / u2.cols;
}


/* function [lines] = getLines(border, gr) */
/* Get lines by position and gradient orientation */
/*
Arguments:
lines      O  3xLEN   homogeneous coordinates of lines fitted to the gradient orientation at given positions
corresp    I  2xLEN   correspondences
gr         I  4(MxN)  gradient (m,a,x,y)
M, N, LEN  I  scalar  image dimensions, number of correspondences
Returns:
Number of orientation reversions
*/
unsigned getLines(double *lines, const unsigned *corresp, const Gradient gr, unsigned LEN, double pointShift) {
	unsigned i;
	unsigned px, py, ix, iy;
	double pt[6], *l = lines;
	double lnorm;
	unsigned reversions = 0;
	unsigned M = gr.m.size().height;
	unsigned N = gr.m.size().width;

	pt[2] = 1;
	pt[5] = 0;
	for (i = 0; i < LEN; ++i) {
		px = corresp[0];
		py = corresp[1];
		pt[0] = px + pointShift;
		pt[1] = py + pointShift;
		pt[3] = pt[4] = 0;
		for (ix = max(0u, px-1); ix <= min(N-1, px+1); ++ix) {
			for (iy = max(0u, py-1); iy <= min(M-1, py+1); ++iy) {
//				pt[4] -= gr.x[(ix-1)*M + iy - 1];
//				pt[3] += gr.y[(ix-1)*M + iy - 1];
				pt[4] -= gr.x.at<double>(iy, ix);
				pt[3] += gr.y.at<double>(iy, ix);
			}
		}
		cross(l, pt, pt+3);
		lnorm = sqrt(l[0]*l[0]+l[1]*l[1]); /* normalisation to unit length of normal vector */
		l[0] /= lnorm;
		l[1] /= lnorm;
		l[2] /= lnorm;

		if (cos(gr.a.at<double>(py, px) - atan2(l[1], l[0])) < 0) { /* Oriented lines */
			++reversions;
			l[0] *= -1;
			l[1] *= -1;
			l[2] *= -1;
		}

		l += 3;
		corresp += 2;
	}
	return reversions;
}

double fastTan1 (double fAngle) {
	double fASqr = fAngle*fAngle;
	double fResult = 9.5168091e-03f;
	fResult *= fASqr;
	fResult += 2.900525e-03f;
	fResult *= fASqr;
	fResult += 2.45650893e-02f;
	fResult *= fASqr;
	fResult += 5.33740603e-02f;
	fResult *= fASqr;
	fResult += 1.333923995e-01f;
	fResult *= fASqr;
	fResult += 3.333314036e-01f;
	fResult *= fASqr;
	fResult += 1.0f;
	fResult *= fAngle;
	return fResult;
}

double fastInvTan1 (double fValue) {
	double fVSqr = fValue*fValue;
	double fResult = 0.0028662257f;
	fResult *= fVSqr;
	fResult -= 0.0161657367f;
	fResult *= fVSqr;
	fResult += 0.0429096138f;
	fResult *= fVSqr;
	fResult -= 0.0752896400f;
	fResult *= fVSqr;
	fResult += 0.1065626393f;
	fResult *= fVSqr;
	fResult -= 0.1420889944f;
	fResult *= fVSqr;
	fResult += 0.1999355085f;
	fResult *= fVSqr;
	fResult -= 0.3333314528f;
	fResult *= fVSqr;
	fResult += 1.0f;
	fResult *= fValue;
	return fResult;
}

double lutTanSimple(double angle) { /* usable when we know the angle is in range [0;pi/2] and the TAN_TABLE is already set */
	return TAN_TABLE[(unsigned)(angle*1000)];
}

double lutTan(double angle) {
	while (angle > M_PI / 2) {
		angle -= M_PI;
	}
	if (angle < 0) {
		return -lutTan(-angle); /* Odd function */
	}
	
	/* Now the angle is in range [0, pi/2] */
	
	if (!tanTableSet) {
		unsigned i;
		tanTableSet = 1;
		for (i = 0; i < 1572; ++i) { /* range 0..pi/2, step 0.001 */
			TAN_TABLE[i] = tan(0.001*i);
		}
	}
	
	return TAN_TABLE[(unsigned)(angle*1000)];
}

double lutATanS(double value) {
	if (value < 0) {
		return -lutATanS(-value); /* Odd function */
	}
	if (value >= 1000) {
		return M_PI / 2;
	} /* Now the value is in range [0, 1000) */
	
	if (value < 10) {
		return ATAN_TABLE[(unsigned)(value*1000)];
	} else if (value < 100) {
		return ATAN_TABLE[10000 + (unsigned)(value*10) - 100];
	} else {
		return ATAN_TABLE[10900 + (unsigned)(value*0.1) - 10];
	}
}

double lutATan(double value) {
	if (value < 0) {
		return -lutATan(-value); /* Odd function */
	}
	if (value >= 1000) {
		return M_PI / 2;
	} /* Now the value is in range [0, 1000) */
	
	if (!atanTableSet) {
		unsigned i;
		atanTableSet = 1;
		for (i = 0; i < 10000; ++i) { /* range 0..10, step 0.001 */
			ATAN_TABLE[i] = atan(0.001*i);
		}
		for (i = 100; i < 1000; ++i) { /* range 10..100, step 0.1 */
			ATAN_TABLE[10000 + i - 100] = atan(0.1 * i);
		}
		for (i = 10; i <= 100; ++i) { /* range 100..1000, step 10 */
			ATAN_TABLE[10900 + i - 10] = atan(10 * i);
		}
	}
	
	if (value < 10) {
		return ATAN_TABLE[(unsigned)(value*1000)];
	} else if (value < 100) {
		return ATAN_TABLE[10000 + (unsigned)(value*10) - 100];
	} else {
		return ATAN_TABLE[10900 + (unsigned)(value*0.1) - 10];
	}
}


double lutInterpTan(double angle) {
	double index;
	unsigned indexH, indexL;
	
	while (angle > M_PI / 2) {
		angle -= M_PI;
	}
	if (angle < 0) {
		return -lutInterpTan(-angle); /* Odd function */
	} /* Now the angle is in range [0, pi/2] */
	
	if (!tanTableSet) {
		unsigned i;
		tanTableSet = 1;
		for (i = 0; i < 1572; ++i) { /* range 0..pi/2, step 0.001 */
			TAN_TABLE[i] = tan(0.001*i);
		}
	}
	
	index = angle*1000;
	indexL = (unsigned)index;
	indexH = indexL + 1;
	return TAN_TABLE[indexL]*(1-(index-indexL)) + TAN_TABLE[indexH]*(1-(indexH-index));
}

double lutInterpATan(double value) {
	double index;
	unsigned indexH, indexL;
	
	if (value < 0) {
		return -lutInterpATan(-value); /* Odd function */
	}
	if (value >= 1000) {
		return M_PI / 2;
	} /* Now the value is in range [0, pi/2] */
	
	if (!atanTableSet) {
		unsigned i;
		atanTableSet = 1;
		for (i = 0; i < 10000; ++i) { /* range 0..10, step 0.001 */
			ATAN_TABLE[i] = atan(0.001*i);
		}
		for (i = 100; i < 1000; ++i) { /* range 10..100, step 0.1 */
			ATAN_TABLE[10000 + i - 100] = atan(0.1 * i);
		}
		for (i = 10; i <= 100; ++i) { /* range 100..1000, step 10 */
			ATAN_TABLE[10900 + i - 10] = atan(10 * i);
		}
	}
	
	if (value < 10) {
		index = value*1000;
	} else if (value < 100) {
		index = 10000 + value*10 - 100;
	} else {
		index = 10900 + value*0.1 - 10;
	}

	indexL = (unsigned)index;
	indexH = indexL + 1;
	return ATAN_TABLE[indexL]*(1-(index-indexL)) + ATAN_TABLE[indexH]*(1-(indexH-index));
}


/* function [ptQuality] = updatePtQuality(ptQualityOld, correspFull, lines, A, ptQualityDec, evid) */
/* Update point quality by warping and adding ones at point positions */
/*
Arguments:
ptQualNew    O   MxN   transformed field of point qualities
ptQualOld    I   MxN   field of point qualities in the previous frame f_t-1
corresp      I  4xLEN  correspondences from f_t-1 to f_t
lines        I  3xLEN  lines in the previous frame f_t-1, homogeneous coordinates
A            I   3x3   transformation from f_t-1 to f_t
ptQualityDec I  scalar fading factor
evidDist     I   MxN   chamfer distance field in f_t
evidAng      I   MxN   angle of closest edge point in f_t
M, N, LEN    I  scalar image dimensions, number of correspondences
*/
void updatePtQuality(double *ptQualNew, const double *ptQualOld, const unsigned *corresp, const double *lines, const double *A,
					double ptQualityDec, const double *evidDist, const double *evidAng, unsigned M, unsigned N, unsigned LEN) {
	
	unsigned x, y, i, xl, yl, iN, iO;
	double xb, yb, vll, vlh, vhl, vhh;
	double angle, lAngle, rot = a2rot(A), Ai[3*3];
	double ptQualW, ptAngleW;
	
	memcpy(Ai, A, 3*3*sizeof(double));
	inv3(Ai);
	
	for (y = 1; y <= M; ++y) { /* coordinates, not indices! */
		for (x = 1; x <= N; ++x) {
			projectA(&xb, &yb, Ai, &x, &y, 1);
			xl = (unsigned) xb; /* lo coordinates for interpolation */
			yl = (unsigned) yb;
			if (xb >= 1 && xb < N && yb >= 1 && yb < M) { /* backprojection fell inside image */
				vll = ptQualOld[(xl-1)*M + yl - 1];
				vlh = ptQualOld[(xl-1)*M + yl];
				vhl = ptQualOld[(xl  )*M + yl - 1];
				vhh = ptQualOld[(xl  )*M + yl];
				ptQualNew[(x-1)*M + y - 1] = /* bilinear interpolation */
						(	vll*(1.0-xb+xl)*(1.0-yb+yl) +
							vlh*(1.0-xb+xl)*(yb-yl) +
							vhl*(xb-xl)*(1.0-yb+yl) +
							vhh*(xb-xl)*(yb-yl) ) *
						ptQualityDec;
			} else {
				ptQualNew[(x-1)*M + y - 1] = 0; /* outside of image, nothing to get */
			}
		}
	}
	
	for (i = 0; i < LEN; ++i) {
		x = corresp[0];
		y = corresp[1];
		projectA(&xb, &yb, A, &x, &y, 1);
		xl = (unsigned) (xb + .5); /* fast round :) */
		yl = (unsigned) (yb + .5);
		iO = (x - 1) * M + y - 1;
		iN = (xl - 1) * M + yl - 1;
		if (xl >= 1 && xl <= N && yl >= 1 && yl <= M) {
			ptQualW = evidDist[iN];
			angle = evidAng[iN]; /* expected new angle */
			lAngle = atan2(lines[1], lines[0]); /* old angle (+rot(A)!) */
			ptAngleW = .5 + .5 * cos(lAngle + rot - angle);
			if (ptQualNew[iN] < ptQualityDec * ptQualOld[iO] + ptQualW * ptAngleW) {
			/* speed-up: isn't this condition always true? Was already warped and now we add only positive values? */
				ptQualNew[iN] = ptQualityDec * ptQualOld[iO] + ptQualW * ptAngleW;
				/* probably should be just += ptQualW * ptAngleW? only difference by interpolation vs. NN? */
			}
		}
		
		corresp += 4;
		lines += 3;
	}
}


// a wrapper for the above
void updatePtQualityWrapper(cv::Mat& ptQuality,
					const std::vector<unsigned> corresp, const std::vector<double> lines, const cv::Mat& A, double ptQualityDec, const Evidence evidence) {
	
	assert(corresp.size() % 4 == 0);
	assert(lines.size() % 3 == 0);
	
	Mat ptQualT = ptQuality.t();
	Mat ptQualNewT = ptQualT.clone();
	Mat AT = A.t();
	Mat evidDistT = evidence.d.t();
	Mat evidAngT = evidence.a.t();
	
	if (!ptQualT.isContinuous() || !ptQualNewT.isContinuous() || !AT.isContinuous() || !evidDistT.isContinuous() || !evidAngT.isContinuous()) {
		throw "Can't get the data from a non-continuous matrix!";
	}
	
	updatePtQuality((double*)ptQualNewT.ptr(), (double*) ptQualT.ptr(), &corresp[0], &lines[0], (double*) AT.ptr(),
					ptQualityDec, (double*) evidDistT.ptr(), (double*) evidAngT.ptr(), ptQuality.rows, ptQuality.cols, corresp.size()/4);
	
	// now we need to save it back!
	ptQuality = ptQualNewT.t();
}



/**
 * getGradient
 *
 * @param Gradient gr       O 4(MxN)  structure with gradient
 * @param const double *img I  MxN    input image, grayscale
 * @param unsigned M        I scalar  image height
 * @param unsigned N        I scalar  image width
 * @param unsigned *limits  I   4     computation limits (xMin, yMin, xMax, yMax)
 *
 * @brief Computes gradient of a gray image -- components, angle and magnitude.
 */
void getGradient(GradientD gr, const double *img, unsigned M, unsigned N, unsigned *limits) {
	unsigned i, j, k, xM, yM;
	const double *imgL = img - M, *imgR = img + M, *imgU = img - 1, *imgD = img + 1;
	double gx, gy;
	
	/* corners */
	if (xMin == 1) {
		if (yMin == 1) {
			gr.x[0] = imgR[0] - img[0];
			gr.y[0] = imgD[0] - img[0];
		}
		if (yMax == M) {
			gr.x[M-1] = imgR[M-1] - img[M-1];
			gr.y[M-1] = img[M-1] - imgU[M-1];
		}
	}
	if (xMax == N) {
		if (yMin == 1) {
			gr.x[(N-1)*M] = img[(N-1)*M] - imgL[(N-1)*M];
			gr.y[(N-1)*M] = imgD[(N-1)*M] - img[(N-1)*M];
		}
		if (yMax == M) {
			gr.x[N*M-1] = img[N*M-1] - imgL[N*M-1];
			gr.x[N*M-1] = img[N*M-1] - imgU[N*M-1];
		}
	}
	
	/* first column */
	if (xMin == 1) {
		for (i = 1; i < M - 1; ++i) {
			gr.x[i] = imgR[i] - img[i];
			gr.y[i] = (imgD[i] - imgU[i]) / 2;
		}
	}
	
	/* last column */
	if (xMax == N) {
		for (i = (N-1)*M + 1; i < N*M - 1; ++i) {
			gr.x[i] = img[i] - imgL[i];
			gr.y[i] = (imgD[i] - imgU[i]) / 2;
		}
	}
	
	/* first row */
	if (yMin == 1) {
		for (i = M; i < (N-1)*M; i += M) {
			gr.x[i] = (imgR[i] - imgL[i]) / 2;
			gr.y[i] = imgD[i] - img[i];
		}
	}
	
	/* last row */
	if (yMax == M) {
		for (i = 2*M - 1; i < N*M - 1; i += M) {
			gr.x[i] = (imgR[i] - imgL[i]) / 2;
			gr.y[i] = img[i] - imgU[i];
		}
	}
	
	/* the rest of image */
	yM = min(M - 1, yMax);
	xM = min(N - 1, xMax);
	for (j = max(1u, yMin - 1); j < yM; ++j) { /* row index */
		for (k = max(1u, xMin - 1); k < xM; ++k) { /* column index */
			i = k*M + j;
			gr.x[i] = (imgR[i] - imgL[i]) / 2;
			gr.y[i] = (imgD[i] - imgU[i]) / 2;
		}
	}
	
	/* magnitude and angle */
	for (j = yMin - 1; j < yMax; ++j) {
		for (k = xMin - 1; k < xMax; ++k) {
			i = k*M + j;
			gx = gr.x[i];
			gy = gr.y[i];
			gr.a[i] = atan2(gy, gx);
			gr.m[i] = sqrt(gx*gx + gy*gy);
		}
	}
	
}




} // namespace ltflo


