#include "ransacTools.h"
#include <string.h>
#include <stdlib.h>
#include <math.h>


namespace ltflo {

/*maximum number of RANSAC's samples*/
#define MAX_SAM (5000u)
/*tolerance of similarity matrix determinant (1% S[3,3])*/
#define DET_TOL (0.01)
/*length of one correspondence vector, 2*3 lines + 2*3 points + index*/
#define U_SIZE (13)
/*Inner RANSAC iterations*/
#define INNER_RSC (5)

/* Lines RANSAC for SIMILARITY estimation
All the matrices are stored column-wise!
output arguments:
A              -o- 3*3    - similarity matrix
inliers        -o- 1*LEN  - 0/1
D              -o- 1*LEN  - errors of correspondences
score          -o- scalar - evidence of returned transf.
input arguments:
u              -i- 13*LEN - correspondences in homogeneous coordinates,
                            lines, points and index together, in form [l1;l2;x1;x2;ind]
                            (Is changed during the computation! Indices sorted by the angle have to be returned.)
evidDist       -i- M*N    - inverted Chamfer map
evidAng        -i- M*N    - angle of nearest edge
winOld         -i- 2*4    - trackedArea for overlap comparison
win            -i- 2*4    - tracked area
thr            -i- scalar - RANSAC's ilier/outlier error threshold
conf           -i- scalar - RANSAC's terminating confidence
expScaleChange -i- scalar - expected change of scale
verbose        -i- scalar - verbose mode
M              -i- scalar - image height
N              -i- scalar - image width
LEN            -i- scalar - number of correspondences

*/
void ransac(double * A, bool * inliers, double * D, double * score,
			double * u, const double * evidDist, const double * evidAng,
			const double * winOld, const double * win, double thr, double conf, double expScaleChange, bool verbose,
			unsigned M, unsigned N, unsigned LEN, bool repair);
			
/*Performs local optimisation according to Chum, Matas, Kittler: Locally Optimized RANSAC, DAGM 2003*/
double localOptimisation(double * A,
			double * d, const double * u, double thr, const double * evidDist, const double * evidAng,
			const double * winOld, const double * win, double expScaleChange,
			unsigned M, unsigned N, unsigned len, double *buffer, bool repair, bool verbose);

/*performs iterative least squares with narrowing threshold from thrs to thr*/
double iterativeLSq(double * A,
			double * d, const double * u, double thr, double thrs, unsigned iters,
			const double * evidDist, const double * evidAng, const double * winOld, const double * win,
			double expScaleChange, unsigned M, unsigned N, unsigned len, double *buffer, bool repair);


// outer thing, includes the ransac and refinement, returns number of inliers
unsigned ransacWrapper(cv::Mat& A, std::vector<double>& d, double& score, double& optScore,
					std::vector<double>& linesOld, std::vector<double>& linesNew, std::vector<unsigned>& correspFull,
					const Evidence& evidence, const cv::Mat& weightsField, const BBox& winOld, const BBox& winNew,
					double errThr, double conf, double scaleChange, bool repair, bool verbose, const cv::Mat& image);



} // namespace ltflo


