#ifndef __RANSAC_TOOLS_H__
#define __RANSAC_TOOLS_H__

#include "algebra.h"
#include "intersection.h"
#include <math.h>
#include <stdlib.h>


namespace ltflo {


//unsigned char REPAIR;

/*Compare angles of l1 modulo PI, both parameters are joint correspondences [l1;l2;l3;l4]*/
int compareAngle(const void * u1, const void * u2);

/*Number of samples to report success of 'ss'-point RANSAC with 'inl' inliers out of 'len' points with confidence 'conf'*/
unsigned nSamples(unsigned len, unsigned inl, unsigned ss, double conf);

/*Generate samples = triplets of lines with large mutual angles*/
void getRandTriangle(unsigned * sample, unsigned len, unsigned noTriangles);

/*fills binary field with 1's at positions of points with error lower than threshold, inlier count returned*/
unsigned getInliers(bool * inliers, double * d, double thr, unsigned len);

/*finds positions of points with error lower than threshold, inlier count returned*/
unsigned getInlierList(unsigned * inliers, double * d, double thr, unsigned len);

/*computes evidence of a transformation*/
double evidenceScore(double * A, const double * u, const double * evidDist, const double * evidAng, double expScaleChange,
					const double * winOld, const double * win,
					unsigned M, unsigned N, unsigned len, unsigned u_len, bool repair);

/*computes a similarity transformation from T triplets of line intersections (triangles vetrices), return error code
error codes:
 0 - success
 1 - degenerate sample*/
int triangle2A(double * A, const double * u, const unsigned * indexes, unsigned noTriangles, unsigned u_len);



} // namespace ltflo


#endif // __RANSAC_TOOLS_H__

